/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */

#include <iostream>
#include <string>
#include <vector>
#include <complex>
#include <math.h>
#include <algorithm>

#include <net_types.h>
#include <Vector.h>
#include <Matrix.h>
#include <Map.h>
#include <Source.h>
#include <BufferedNode.h>

////#include <boost/python.hpp>
////#include <boost/python/numeric.hpp>
////#include <boost/python/suite/indexing/vector_indexing_suite.hpp>
////#include <boost/python/suite/indexing/map_indexing_suite.hpp>
//#include <pybind11/eval.h>
//#include <pybind11/pybind11.h>
#include <pybind11/embed.h>
#include <pybind11/stl.h>
#include <pybind11/operators.h>
#include <pybind11/numpy.h>
//#include <pybind11/eigen.h>

////#include <boost/uuid/uuid.hpp>
////#include <boost/uuid/uuid_io.hpp>
////#include <boost/uuid/uuid_generators.hpp>

////namespace bp = boost::python;
namespace py = pybind11;
////namespace bu = boost::uuids;
using namespace FD;
////using namespace std;

class PyCodeExecutor3;

DECLARE_NODE(PyCodeExecutor3);
/*Node
 *
 * @name PyCodeExecutor3
 * @category HARK:Python
 * @description This node executs an external python code. 
 *
 * @parameter_name DIRECTORY_NAME
 * @parameter_type string
 * @parameter_value .
 * @parameter_description [optional] The directory name of your python code. It will inserted to sys.path.
 *
 * @parameter_name MODULENAME
 * @parameter_type string
 * @parameter_value samplecode
 * @parameter_description Your python module name to import, i.e., your python code file name WITHOUT extension.
 * 
 * @parameter_name CLASSNAME
 * @parameter_type string
 * @parameter_value HarkNode
 * @parameter_description Your class name to call in this node.
 * 
 * @parameter_name DEBUG
 * @parameter_type bool
 * @parameter_list true:false
 * @parameter_value false
 * @parameter_description Debug option. If true, it prints input list and output list.
 * 
 END*/

class PyCodeExecutor3 : public BufferedNode {
  std::vector<int> inputsID;
  std::vector<std::string> inputsName; 

  std::vector<int> outputsID;
  std::vector<std::string> outputsName;
  std::vector<std::string> outputsType;

  ////bp::object main_namespace;
  py::object main_namespace;
  std::string myNodeName;
  std::string classname;
  std::string filename;

  bool isDebug;
  bool isCalculated;
  std::string uuid;

  void uuidreplace(std::string& str, const std::string& from, const std::string& to){
    std::string::size_type pos = 0;
    while(pos = str.find(from, pos), pos != std::string::npos){
      str.replace(pos, from.length(), to);
      pos += to.length();
    }
  }

public:
  PyCodeExecutor3(std::string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    // get parameters
    std::string dirname = "";
    if(parameters.exist("DIRECTORY_NAME"))
      dirname = object_cast<std::string>(parameters.get("DIRECTORY_NAME"));
    filename = object_cast<std::string>(parameters.get("MODULENAME"));
    classname = object_cast<std::string>(parameters.get("CLASSNAME"));
    isDebug = dereference_cast<bool>(parameters.get("DEBUG"));
    isCalculated = false;

    // generate unique id
    ////uuid = boost::lexical_cast<std::string>(bu::random_generator()());
    ////uuidreplace(uuid, "-", "_");
    ////if (isDebug) std::cerr << "[log] UUID:" << uuid << std::endl;

    //Py_Initialize();
    if (!Py_IsInitialized())
      py::initialize_interpreter();
    try{
      ////bp::object main_module = bp::import("__main__");
      ////main_namespace = main_module.attr("__dict__");
      main_namespace = py::dict(py::module::import("__main__").attr("__dict__"));

      ////bp::numeric::array::set_module_and_type("numpy", "ndarray");
      
      // get nodeID and name
      main_namespace["uuid"] = py::module::import("uuid");
      uuid = py::eval("str(uuid.uuid4())", main_namespace).cast<std::string>();
      uuidreplace(uuid, "-", "_");
      if (isDebug) std::cerr << "[log] UUID:" << uuid << std::endl;

      myNodeName = nodeName + uuid;
      int nodeID = -1;
      std::stringstream ss(nodeName.substr(nodeName.rfind("_")+1));
      ss >> nodeID;
      if (isDebug) std::cerr <<  "[log] importing modules and numpy." << std::endl;
      ////main_namespace["numpy"] = bp::import("numpy");
      ////main_namespace["pprint"] = bp::import("pprint");
      main_namespace["numpy"] = py::module::import("numpy");
      main_namespace["pprint"] = py::module::import("pprint");
      main_namespace["harkbasenode"] = py::module::import("harkpython.harkbasenode");
      
      // add the path for user's script
      ////main_namespace["sys"] = bp::import("sys");
      ////if(dirname.length() > 0){
      ////  bp::list path = bp::extract<bp::list>(main_namespace["sys"].attr("path"));
      ////  path.append(dirname.c_str());
      ////}
      main_namespace["sys"] = py::module::import("sys");
      if(dirname.length() > 0){
        // py::extract<T> may be necessary for conversion from py::object.
        py::list path = main_namespace["sys"].attr("path"); // py::extract<bp::list>(...)
        path.append(py::cast(dirname.c_str()));
      }
      ////if(isDebug){ exec("print 'sys.path'; pprint.pprint(sys.path)", main_namespace); }
      py::exec("print(sys.version_info)", main_namespace); // FOR DEBUG
      if(isDebug){
        py::exec("print(sys.path)\npprint.pprint(sys.path)", main_namespace);
      }

      if(isDebug){ std::cerr <<  "[log] importing harkpython." << std::endl; }
      ////exec("from harkpython import *", main_namespace);
      py::exec("from harkpython import *", main_namespace);
      if(isDebug){ std::cerr <<  "[log] importing " << filename << "." << std::endl; }
      ////main_namespace[filename.c_str()] = bp::import(filename.c_str());
      main_namespace[filename.c_str()] = py::module::import(filename.c_str());

      // instantiate user's code.
      main_namespace[myNodeName.c_str()] 
      	= main_namespace[filename.c_str()].attr(classname.c_str())();
      main_namespace[myNodeName.c_str()].attr("nodeID") = nodeID;
      main_namespace[myNodeName.c_str()].attr("nodeName") = nodeName;
      main_namespace[myNodeName.c_str()].attr("makeup")();

      // tell python interpreter how to deal with STL obects.
/*      if (main_namespace.attr("has_key")("convertersRegistered") == false){
        bp::class_<std::vector<complex<float> > > ("PyVecComplexFloat")
          .def(bp::vector_indexing_suite<std::vector<complex<float> > >());
        bp::class_<std::vector<int> > ("PyVecInt")
          .def(bp::vector_indexing_suite<std::vector<int> >());
        bp::class_<std::vector<float> > ("PyVecFloat")
          .def(bp::vector_indexing_suite<std::vector<float> >());
        bp::class_<map<int, std::vector<int> > > ("PyMapVecInt")
          .def(bp::map_indexing_suite<map<int, std::vector<int> > >());
        bp::class_<map<int, std::vector<float> > > ("PyMapVecFloat")
          .def(bp::map_indexing_suite<map<int, std::vector<float> > >());
        bp::class_<map<int, std::vector<complex<float> > > > ("PyMapVecComplex")
          .def(bp::map_indexing_suite<map<int, std::vector<complex<float> > > >());
        bp::class_<map<int, float > > ("PyMapPrimFloat")
          .def(bp::map_indexing_suite<map<int, float > >());

        bp::to_python_converter<Vector<int>, Vector2py<int> >();
        bp::to_python_converter<Vector<float>, Vector2py<float> >();
        bp::to_python_converter<Vector<complex<float> >, Vector2py<complex< float> > >();
        bp::to_python_converter<Matrix<int>, Matrix2py<int> >();
        bp::to_python_converter<Matrix<float>, Matrix2py<float> >();
        bp::to_python_converter<Matrix<complex<float> >, Matrix2py<complex<float> > >();
        bp::to_python_converter<Int, Prim2py<int> >();
        bp::to_python_converter<Float, Prim2py<float> >();
        bp::to_python_converter<Complex<float>, ComplexPrim2py<float> >();
        bp::to_python_converter<Complex<double>, ComplexPrim2py<double> >();
        bp::to_python_converter<Complex<int>, ComplexPrim2py<int> >();
        bp::to_python_converter<Source, Source2py>();
        bp::to_python_converter<RCPtr< Map<int, RCPtr<Vector<int> > > >, Map2py<int> >();
        bp::to_python_converter<RCPtr< Map<int, RCPtr<Vector<float> > > >, Map2py<float> >();
        bp::to_python_converter<RCPtr< Map<int, RCPtr<Vector<complex<float> > > > >, Map2py<complex<float> > >();

        // for PARAM_MAP
        //bp::to_python_converter<RCPtr< Map<std::string, RCPtr< Map <std::string, ObjectRef> > > >, MapMap2py >();

        main_namespace["convertersRegistered"] = true;
      }*/
    }
    ////catch(bp::error_already_set){
    catch(const py::error_already_set& e){
      std::cerr << e.what() << std::endl;
      ////PyErr_Print();
      throw (new NodeException(this, "Error in constructor", __FILE__, __LINE__));
    }
    inOrder = true;
  }

/*  template <typename T>
  struct Map2py{
    static PyObject* convert(RCPtr<Map<int, RCPtr<Vector<T> > > > const& s){
      ////bp::dict topy = bp::dict();
      py::dict topy = py::dict();
      typename Map<int, RCPtr<Vector<T> > >::const_iterator it = s->begin();
      typename Map<int, RCPtr<Vector<T> > >::const_iterator ed = s->end();
      for(; it != ed; it++) topy[it->first] = object_cast<Vector<T> >((it->second));
      return bp::incref(bp::object(topy).ptr());
    }
  };*/
/*
  struct MapMap2py{
    static PyObject* convert(RCPtr< Map<std::string, RCPtr< Map <std::string, ObjectRef> > > > const& s){
      ////bp::dict topy = bp::dict();
      py::dict topy = py::dict();
      typename Map<std::string, RCPtr< Map <std::string, ObjectRef> > >::const_iterator it1 = s->begin();
      typename Map<std::string, RCPtr< Map <std::string, ObjectRef> > >::const_iterator ed1 = s->end();
      for(; it1 != ed1; it1++){
        ////bp::dict map1 = bp::dict();
        py::dict map1 = py::dict();
        typename  Map <std::string, ObjectRef>::const_iterator it2 = (it1->second)->begin();
        typename  Map <std::string, ObjectRef>::const_iterator ed2 = (it1->second)->end();
        for(; it2 != ed2; it2++){
            if(typeid(Int) == typeid(*(it2->second))){
              RCPtr<Int> tmp = it2->second;
              map1[it2->first] = *tmp;
            } else if(typeid(Float) == typeid(*(it2->second))){
              RCPtr<Float> tmp = it2->second;
              map1[it2->first] = *tmp;
            } else if(typeid(String) == typeid(*(it2->second))){
              RCPtr<String> tmp = it2->second;
              map1[it2->first] = *tmp;
            }
        }

        topy[it1->first] = map1;
      }
      return bp::incref(bp::object(topy).ptr());
    }
  };
*/
/*  template <typename T>
  struct Prim2py{
    static PyObject* convert(NetCType<T> const& s) {
      NetCType<T> &topy = const_cast< NetCType < T >  & > ( s );
      return bp::incref(bp::object(topy.val()).ptr());
    }
  };
  template <typename T>
  struct ComplexPrim2py{
    static PyObject* convert(Complex<T> const& s){
      Complex<T> &topy = const_cast<Complex<T> &> (s);
      return bp::incref(bp::object(topy.val()).ptr());
    }
  };
  template <typename T>
  struct Vector2py{
    static PyObject* convert(Vector<T> const& s){
      return bp::incref(bp::object(std::vector<T>(s)).ptr());
    }
  };

  template <typename T>
  struct Matrix2py{
    static PyObject* convert(Matrix<T> const& input){
      std::vector<T> data(input.nrows() * input.ncols());
      for(int row=0; row<input.nrows(); row++)
	for(int col=0; col<input.ncols(); col++)
	  data[row * input.ncols() + col] = input(row, col);

      bp::numeric::array topy(data);
      ////topy.resize(bp::make_tuple(input.nrows(), input.ncols()));
      topy.resize(py::make_tuple(input.nrows(), input.ncols()));

      return bp::incref(bp::object(topy).ptr());
    }
  };


  struct Source2py{
    static PyObject* convert(Source const& s){
      ////bp::dict topy = bp::dict();
      ////Source& src = const_cast<Source&> (s);
      ////topy["id"] = src.id;
      ////topy["x"] = bp::list();
      ////topy["x"].attr("append")(src.x[0]);
      ////topy["x"].attr("append")(src.x[1]);
      ////topy["x"].attr("append")(src.x[2]);
      ////topy["power"] = src.power;
      ////return bp::incref(topy.ptr());	
      py::dict topy = py::dict();
      Source& src = const_cast<Source&> (s);
      topy["id"] = py::cast(src.id);
      topy["x"] = py::list();
      topy["x"].append(py::cast(src.x[0]));
      topy["x"].append(py::cast(src.x[1]));
      topy["x"].append(py::cast(src.x[2]));
      topy["power"] = py::cast(src.power);
      return bp::incref(topy.ptr());	
    }
  };
  //// pybind11 not need convert function.
  //// maybe we need replace to pybind11::detail::type_caster<T> overload.
  //// type_caster's cast() method (C++ => Python)
  //// and load() method (Python => C++) used.
*/
  ~PyCodeExecutor3(){
    ////try{
    ////  main_namespace[myNodeName.c_str()].attr("__del__")();
    ////} catch(bp::error_already_set){
    try{
      main_namespace[myNodeName.c_str()].attr("__del__")();
    } catch(const py::error_already_set& e){
      if(isCalculated){
        std::cerr << e.what() << std::endl;
      	////PyErr_Print();
      	throw(new NodeException(this, "Error in destructor", __FILE__, __LINE__));      
      } else{}
    }
    //py::finalize_interpreter();
  }
  
  virtual int translateInput(const std::string &inputName){
    int id = addInput(inputName);
    inputsName.push_back(inputName);
    inputsID.push_back(id);

    if(isDebug){
      std::cout << "translate Input (" << myNodeName << ":" << classname << ") "
                << inputName << " (ID: " << id << ")" << std::endl;
    }
    return id;
  }
  
  virtual int translateOutput(const std::string &outputName){
    std::vector<std::string>::iterator result 
      = find(outputsName.begin(), outputsName.end(), outputName);
    
    if(result != outputsName.end()){
      if(isDebug){
        std::cout << outputName << " already exists: " 
                  << result - outputsName.begin() << std::endl;
      }
      return outputsID[result - outputsName.begin()];
    }
    int id = addOutput(outputName);
    outputsName.push_back(outputName);
    outputsID.push_back(id);

    if(isDebug){
      std::cout << "translate Output: (" << myNodeName << "::" << classname << ")" 
                << " Name: " << outputName << " ID: " << id << std::endl;
    }
    ////bp::tuple nam;
    ////bp::tuple typ;
    ////try{
    ////  nam = bp::extract<bp::tuple>(main_namespace[myNodeName.c_str()].attr("outputNames"));
    ////  typ = bp::extract<bp::tuple>(main_namespace[myNodeName.c_str()].attr("outputTypes"));
    ////} catch(bp::error_already_set){
    py::tuple nam;
    py::tuple typ;
    try{
      // py::extract<T> may be necessary for conversion from py::object.
      nam = main_namespace[myNodeName.c_str()].attr("outputNames"); // py::extract<py::tuple>(...)
      typ = main_namespace[myNodeName.c_str()].attr("outputTypes"); // py::extract<py::tuple>(...)
    } catch(const py::error_already_set& e){
      std::cerr << e.what() << std::endl;
      ////PyErr_Print();
      std::stringstream ss;
      ss << "Error during translate Output (" << filename << "::" << myNodeName << "::" << classname << ")"
          << "outputNames or outputTypes is not defined as tuple" << std::endl;
      throw (new NodeException(this, ss.str(),  __FILE__, __LINE__));
    }

    int inam = 0;
    ////try{
    ////  inam = bp::extract<int>(nam.attr("index")(outputName));
    ////} catch(bp::error_already_set){
    try{
      inam = nam.attr("index")(outputName).cast<int>();
    } catch(const py::error_already_set& e){
      std::cerr << e.what() << std::endl;
      ////PyErr_Print();
      std::stringstream ss;
      ss << "Error during translate Output (" << filename << "::" << myNodeName << "::" << classname << ")"
          << " Output Name: " << outputName << " of ID: " << id << " is not found in outputNames." << std::endl
          << "candidates are:::" << std::endl;
      ////for(int i=0; i<bp::len(nam); i++){
      ////  std::string str =  bp::extract<std::string>(nam[i]);
      ////  ss << str << std::endl;
      ////}
      for(size_t i=0; i<nam.size(); i++){
        std::string str =  nam[i].cast<std::string>();
        ss << str << std::endl;
      }
      throw (new NodeException(this, ss.str(),  __FILE__, __LINE__));
    }
    ////outputsType.push_back((std::string) bp::extract<std::string>(typ[inam]));
    outputsType.push_back(typ[inam].cast<std::string>());

    return id;
  }
  
  virtual void initialize(void)
  {
    std::vector<int>::iterator id;

    int maxahead = 0, maxback = 0;
    for(id = outputsID.begin(); id != outputsID.end(); id++){
      if(maxahead < outputs[*id].lookAhead) maxahead = outputs[*id].lookAhead;
      if(maxback < outputs[*id].lookBack) maxback = outputs[*id].lookBack;
    }
    for(id = outputsID.begin(); id != outputsID.end(); id++){
      outputs[*id].lookAhead = maxahead + 1;
      outputs[*id].lookBack = maxback + 1;
    }
    this->BufferedNode::initialize();
  }

  ////template<typename T>
  ////void givePython(RCPtr<T> input, std::string inputName){
    ////main_namespace[myNodeName.c_str()].attr(inputName.c_str()) = *input;
  ////}
  template<typename T>
  py::array_t<T> getNumpyVector(ObjectRef& input) {
    std::vector<T> o = object_cast<std::vector<T>>(input);
    return py::array_t<T>({o.size()}, {sizeof(T)}, o.data());
  }

  template<typename T>
  py::array_t<T> getNumpyMatrix(ObjectRef& input) {
    Matrix<T> o = object_cast<Matrix<T>>(input);
    return py::array_t<T>({o.nrows(), o.ncols()}, {sizeof(T)*o.ncols(), sizeof(T)}, o[0]);
  }


  void calculate(int output_id, int count, Buffer &out)
  {
    // get input
    std::vector<std::string>::iterator inputsNameit = inputsName.begin();
    std::vector<int>::iterator inputsIDit = inputsID.begin();
    try{
      for(; inputsIDit < inputsID.end(); inputsIDit++, inputsNameit++){
        ObjectRef input = getInput(*inputsIDit, count);
        if(isDebug)
          std::cout << "Input terminal (" << myNodeName << "::" << classname << ")  "
               << "Name: " << *inputsNameit << ", ID: " << *inputsIDit << std::endl;

        // give data to python depending on their types 
        if(typeid(Int) == typeid(*input))
          ////givePython(RCPtr<Int>(input), *inputsNameit);
          main_namespace[myNodeName.c_str()].attr(inputsNameit->c_str()) = dereference_cast<int>(input);
        else if(typeid(Float) == typeid(*input))
          ////givePython(RCPtr<Float>(input), *inputsNameit);
          main_namespace[myNodeName.c_str()].attr(inputsNameit->c_str()) = dereference_cast<float>(input);
        else if(typeid(Complex<float>) == typeid(*input)) {
          ////givePython(RCPtr<Complex<float> >(input), *inputsNameit);
          main_namespace[myNodeName.c_str()].attr(inputsNameit->c_str()) = object_cast<std::complex<float>>(input);
        } else if(typeid(Source) == typeid(*input)) {
          ////givePython(RCPtr<Source> (input), *inputsNameit);
          Source src = object_cast<Source>(input);
          py::dict pySource = py::dict();
          pySource["id"] = src.id;
          std::array<float, 3> arrX = {src.x[0], src.x[1], src.x[2]};
          pySource["x"] = arrX;
          pySource["power"] = src.power;
          main_namespace[myNodeName.c_str()].attr(inputsNameit->c_str()) = pySource;
        } else if (typeid(String) == typeid(*input))
          main_namespace[myNodeName.c_str()].attr(inputsNameit->c_str()) = object_cast<std::string>(input);
        else if(typeid(Vector<int>) == typeid(*input))
          ////givePython(RCPtr<Vector<int> >(input), *inputsNameit);
          main_namespace[myNodeName.c_str()].attr(inputsNameit->c_str()) = getNumpyVector<int>(input);
          //main_namespace[myNodeName.c_str()].attr(inputsNameit->c_str()) = object_cast<std::vector<int>>(input);
        else if(typeid(Vector<float>) == typeid(*input))
          ////givePython(RCPtr<Vector<float> >(input), *inputsNameit);
          main_namespace[myNodeName.c_str()].attr(inputsNameit->c_str()) = getNumpyVector<float>(input);
          //main_namespace[myNodeName.c_str()].attr(inputsNameit->c_str()) = object_cast<std::vector<float>>(input);
        else if(typeid(Vector<std::complex<float> >) == typeid(*input))
          ////givePython(RCPtr<Vector<std::complex<float> > >(input), *inputsNameit);
          main_namespace[myNodeName.c_str()].attr(inputsNameit->c_str()) = getNumpyVector<std::complex<float>>(input);
          //main_namespace[myNodeName.c_str()].attr(inputsNameit->c_str()) = object_cast<std::vector<std::complex<float>>>(input);
        else if(typeid(Matrix<float>) == typeid(*input))
          ////givePython(RCPtr<Matrix<float> >(input), *inputsNameit);
          main_namespace[myNodeName.c_str()].attr(inputsNameit->c_str()) = getNumpyMatrix<float>(input);
          //main_namespace[myNodeName.c_str()].attr(inputsNameit->c_str()) = convPythonMatrix(object_cast<Matrix<float>>(input));
        else if(typeid(Matrix<int>) == typeid(*input))
          ////givePython(RCPtr<Matrix<int> >(input), *inputsNameit);
          main_namespace[myNodeName.c_str()].attr(inputsNameit->c_str()) = getNumpyMatrix<int>(input);
          //main_namespace[myNodeName.c_str()].attr(inputsNameit->c_str()) = convPythonMatrix(object_cast<Matrix<int>>(input));
        else if(typeid(Matrix<std::complex<float> >) == typeid(*input))
          ////givePython(RCPtr<Matrix<std::complex<float> > >(input), *inputsNameit);
          main_namespace[myNodeName.c_str()].attr(inputsNameit->c_str()) = getNumpyMatrix<std::complex<float>>(input);
          //main_namespace[myNodeName.c_str()].attr(inputsNameit->c_str()) = convPythonMatrix(object_cast<Matrix<std::complex<float>>>(input));
        else if(typeid(Vector<ObjectRef>) == typeid(*input)) {
          Vector<ObjectRef> vec = object_cast<Vector<ObjectRef> > (input);
          ////bp::list topylist = bp::list();
          py::list topylist = py::list(vec.size());
          for(unsigned int i = 0; i < vec.size(); i++){
            if(typeid(Source) == typeid(*(vec[i]))) {
              ////topylist.append(object_cast<Source>(vec[i]));
              Source src = object_cast<Source>(vec[i]);
              py::dict pySource = py::dict();
              pySource["id"] = src.id;
              std::array<float, 3> arrX = {src.x[0], src.x[1], src.x[2]};
              pySource["x"] = arrX;
              pySource["power"] = src.power;
              topylist[i]=pySource;
            } else {
              std::stringstream ss;
              ss << "Input of Vector<ObjectRef> [ObjectRef = " << typeid(*(vec[i])).name() << "] is not supported";
              throw new NodeException(this, ss.str().c_str(), __FILE__, __LINE__);
            }
          }
          main_namespace[myNodeName.c_str()].attr((*inputsNameit).c_str()) = topylist;
        }
        else if(typeid(*input) == typeid(Map<int, ObjectRef>)) {
          Map<int, ObjectRef>  topy = object_cast<Map<int, ObjectRef> > (input);
          ////bp::dict topydict = bp::dict();

          Map<int, ObjectRef >::iterator it = topy.begin();
          if (typeid(*(it->second)) == typeid(Vector<int>)) {
            auto topymap = std::map<int, py::array_t<int>>();
            for(; it != topy.end(); it++) {
              topymap.emplace_hint(topymap.end(), it->first, getNumpyVector<int>(it->second));
            }
            main_namespace[myNodeName.c_str()].attr((*inputsNameit).c_str()) = topymap;
          } else if(typeid(*(it->second)) == typeid(Vector<float>)) {
            auto topymap = std::map<int, py::array_t<float>>();
            for(; it != topy.end(); it++) {
              topymap.emplace_hint(topymap.end(), it->first, getNumpyVector<float>(it->second));
            }
            main_namespace[myNodeName.c_str()].attr((*inputsNameit).c_str()) = topymap;
          } else if(typeid(*(it->second)) == typeid(Vector<std::complex<float>>)) {
            auto topymap = std::map<int, py::array_t<std::complex<float>>>();
            for(; it != topy.end(); it++) {
              topymap.emplace_hint(topymap.end(), it->first, getNumpyVector<std::complex<float>>(it->second));
            }
            main_namespace[myNodeName.c_str()].attr((*inputsNameit).c_str()) = topymap;
          } else if(typeid(*(it->second)) == typeid(Float)) {
            auto topymap = std::map<int, float>();
            for(; it != topy.end(); it++) {
              topymap.emplace_hint(topymap.end(), it->first, dereference_cast<float>(it->second));
            }
            main_namespace[myNodeName.c_str()].attr((*inputsNameit).c_str()) = topymap;
          } else if(typeid(*(it->second)) == typeid(Matrix<float> )) {
            auto topymap = std::map<int, py::array_t<float>>();
            for(; it != topy.end(); it++) {
              topymap.emplace_hint(topymap.end(), it->first, getNumpyMatrix<float>(it->second));
            }
            main_namespace[myNodeName.c_str()].attr((*inputsNameit).c_str()) = topymap;
          } else if(typeid(*(it->second)) == typeid(Matrix<int> )) {
            auto topymap = std::map<int, py::array_t<int>>();
            for(; it != topy.end(); it++) {
              topymap.emplace_hint(topymap.end(), it->first, getNumpyMatrix<int>(it->second));
            }
            main_namespace[myNodeName.c_str()].attr((*inputsNameit).c_str()) = topymap;
          } else if(typeid(*(it->second)) == typeid(Matrix<std::complex<float> > )) {
            auto topymap = std::map<int, py::array_t<std::complex<float>>>();
            for(; it != topy.end(); it++) {
              topymap.emplace_hint(topymap.end(), it->first, getNumpyMatrix<std::complex<float>>(it->second));
            }
            main_namespace[myNodeName.c_str()].attr((*inputsNameit).c_str()) = topymap;
          } else{
            std::stringstream ss;
            ss << "Map value" << typeid(*(it->second)).name() << " is not supported";
            throw new NodeException(this, ss.str().c_str(), __FILE__, __LINE__);
          }
          /*for(; it != topy.end(); it++){
            if(typeid(*(it->second)) == typeid(Vector<int>))
              ////topydict[it->first] = object_cast<Vector<int> > (it->second);
              topydict[std::to_string(it->first).c_str()] = object_cast<std::vector<int> > (it->second);
            else if(typeid(*(it->second)) == typeid(Vector<float>))
              ////topydict[it->first] = object_cast<Vector<float> > (it->second);
              topydict[std::to_string(it->first).c_str()] = object_cast<std::vector<float> > (it->second);
            else if(typeid(*(it->second)) == typeid(Vector<std::complex<float> >))
              ////topydict[it->first] = object_cast<Vector<std::complex<float> > > (it->second);
              topydict[std::to_string(it->first).c_str()] = object_cast<std::vector<std::complex<float> > > (it->second);
            else if(typeid(*(it->second)) == typeid(Float))
              ////topydict[it->first] = object_cast<Float> (it->second);
              topydict[std::to_string(it->first).c_str()] = dereference_cast<float> (it->second);
            else if(typeid(*(it->second)) == typeid(Matrix<float> ))
              ////topydict[it->first] = object_cast<Matrix<float> > (it->second);
              topydict[std::to_string(it->first).c_str()] = convPythonMatrix(object_cast<Matrix<float>>(input));
            else if(typeid(*(it->second)) == typeid(Matrix<int> ))
              ////topydict[it->first] = object_cast<Matrix<int> > (it->second);
              topydict[std::to_string(it->first).c_str()] = convPythonMatrix(object_cast<Matrix<int>>(input));
            else if(typeid(*(it->second)) == typeid(Matrix<std::complex<float> > ))
              ////topydict[it->first] = object_cast<Matrix<std::complex<float> > > (it->second);
              topydict[std::to_string(it->first).c_str()] = convPythonMatrix(object_cast<Matrix<std::complex<float>>>(input));
            else{
              std::stringstream ss;
              ss << "Map value" << typeid(*(it->second)).name() << " is not supported";
              throw new NodeException(this, ss.str().c_str(), __FILE__, __LINE__);
            }
          }
          main_namespace[myNodeName.c_str()].attr((*inputsNameit).c_str()) = topydict;*/
        }
        else if(typeid(*input) == typeid(Map<std::string, ObjectRef>)){
          Map<std::string, ObjectRef>  topy = object_cast<Map<std::string, ObjectRef> > (input);
          ////bp::dict topydict = bp::dict();
          py::dict topydict = py::dict();

          Map<std::string, ObjectRef >::iterator it = topy.begin();
          for(; it != topy.end(); it++){
            if(typeid(*(it->second)) == typeid(Map<std::string, int>))
              topydict[it->first.c_str()] = object_cast<Int> (it->second);
            else if(typeid(*(it->second)) == typeid(Map<std::string, float>))
              topydict[it->first.c_str()] = object_cast<Int> (it->second);
            else if(typeid(*(it->second)) == typeid(Map<std::string, std::string>))
              topydict[it->first.c_str()] = object_cast<String> (it->second);
            else{
              std::stringstream ss;
              ss << "Map<std::string, T> value" << typeid(*(it->second)).name() << " is not supported";
              throw new NodeException(this, ss.str().c_str(), __FILE__, __LINE__);
            }
          }
          main_namespace[myNodeName.c_str()].attr((*inputsNameit).c_str()) = topydict; 
        }
        else{
          if(isDebug) std::cout << std::endl;
          std::stringstream ss;
          ss << "Type (" << typeid(*input).name() << ") of input terminal "
             << *inputsNameit << " is unsupported." << std::endl;
          throw new NodeException (this, ss.str(), __FILE__, __LINE__); 
        }
      }
    ////} catch(bp::error_already_set){
    } catch(const py::error_already_set& e){
      std::cerr << e.what() << std::endl;
      ////PyErr_Print();
      throw (new NodeException(this, "Error during getting inputs", __FILE__, __LINE__));
    }
    
    if(isDebug){
      std::cerr << "calculating..." << std::endl;
    }
    //calculate
    try{
      main_namespace[myNodeName.c_str()].attr("count") = count;
      ////exec((myNodeName + ".calculate()").c_str(), main_namespace);
      py::exec((myNodeName + ".calculate()"), main_namespace);
    }
    ////catch(bp::error_already_set){
    catch(const py::error_already_set& e){
      std::cerr << e.what() << std::endl;
      ////PyErr_Print();
      throw(new NodeException(this, "Error during calculate function execution", __FILE__, __LINE__));
    }
    
    //set output
    std::vector<std::string>::iterator outputsNameit = outputsName.begin();
    std::vector<int>::iterator outputsIDit = outputsID.begin();
    std::vector<std::string>::iterator outputsTypeit = outputsType.begin();
    /*try{
      main_namespace[myNodeName.c_str()].attr("convertValues")();
    }
    ////catch(bp::error_already_set){
    catch(const py::error_already_set& e){
      std::cerr << e.what() << std::endl;
      ////PyErr_Print();
      throw (new NodeException(this, "converting values failed",  __FILE__, __LINE__));
    }*/
    try{
      ////bp::dict d = bp::extract<bp::dict>(main_namespace[myNodeName.c_str()].attr("outputValues"));
      py::dict d = main_namespace[myNodeName.c_str()].attr("outputValues");
      
      for(; outputsIDit < outputsID.end(); outputsTypeit++, outputsNameit++, outputsIDit++){

        if (isDebug)
          std::cout << "Output terminal (" << myNodeName << "::" << classname << ")  " << "Name: " << *outputsNameit
              << ", Type: " << *outputsTypeit << ", ID: " << *outputsIDit << std::endl;

        if(*outputsTypeit == "prim_float"){
          ////(*(outputs[*outputsIDit].buffer))[count] = 
          ////  (new Float(bp::extract<float>(d.get(*outputsNameit))));;
          (*(outputs[*outputsIDit].buffer))[count] = 
            (new Float(d[(*outputsNameit).c_str()].cast<float>()));
        }
        else if(*outputsTypeit == "prim_int"){
          ////(*(outputs[*outputsIDit].buffer))[count] = 
          ////  (new Int(bp::extract<int>(d.get(*outputsNameit))));
          (*(outputs[*outputsIDit].buffer))[count] = 
            (new Int(d[(*outputsNameit).c_str()].cast<int>()));
        }
        else if(*outputsTypeit == "prim_bool"){
          ////(*(outputs[*outputsIDit].buffer))[count] = 
          ////  (new Bool(bp::extract<bool>(d.get(*outputsNameit))));
          (*(outputs[*outputsIDit].buffer))[count] = 
            (new Bool(d[(*outputsNameit).c_str()].cast<bool>()));
        }
        else if(*outputsTypeit == "prim_complex"){
          ////(*(outputs[*outputsIDit].buffer))[count] = 
          ////  (new Complex<float >(bp::extract<complex<float> >(d.get(*outputsNameit))));
          (*(outputs[*outputsIDit].buffer))[count] = 
            (new Complex<float >(d[(*outputsNameit).c_str()].cast<std::complex<float> >()));
        }
        else if(*outputsTypeit == "prim_bool"){
          ////(*(outputs[*outputsIDit].buffer))[count] = 
          ////  (new Bool(bp::extract<bool>(d.get(*outputsNameit))));
          (*(outputs[*outputsIDit].buffer))[count] = 
            (new Bool(d[(*outputsNameit).c_str()].cast<bool>()));
        }
        else if(*outputsTypeit == "prim_source"){
          ////RCPtr<Source> src = RCPtr<Source>(new Source);
          ////bp::dict srcd = bp::extract<bp::dict>(d.get(*outputsNameit));
          ////src->id = bp::extract<int>(srcd.get("id"));
          ////src->x[0] = bp::extract<float>(srcd.get("x")[0]);
          ////src->x[1] = bp::extract<float>(srcd.get("x")[1]);
          ////src->x[2] = bp::extract<float>(srcd.get("x")[2]);
          ////src->power = bp::extract<float>(srcd.get("power"));
          ////(*(outputs[*outputsIDit].buffer))[count] = src;
          RCPtr<Source> src = RCPtr<Source>(new Source);
          py::dict srcd = d[(*outputsNameit).c_str()];
          src->id = srcd["id"].cast<int>();
          src->x[0] = ((py::list)srcd["x"])[0].cast<float>();
          src->x[1] = ((py::list)srcd["x"])[1].cast<float>();
          src->x[2] = ((py::list)srcd["x"])[2].cast<float>();
          src->power = srcd["power"].cast<float>();
          (*(outputs[*outputsIDit].buffer))[count] = src;
        } else if (*outputsTypeit == "prim_string")
          (*(outputs[*outputsIDit].buffer))[count] = (new String(d[(*outputsNameit).c_str()].cast<std::string>()));
        else if(*outputsTypeit == "vector_int"){
          ////bp::list tmplist = bp::extract<bp::list>(d.get(*outputsNameit));
          ////RCPtr<Vector<int> > frompy(new Vector<int>());
          ////for(int i=0; i<bp::len(tmplist); i++){
          ////  (*frompy).push_back((int)bp::extract<int>(tmplist[i]));
          ////}
          ////(*(outputs[*outputsIDit].buffer))[count] = frompy;
          py::list tmplist = d[(*outputsNameit).c_str()];
          RCPtr<Vector<int> > frompy(new Vector<int>());
          for(size_t i=0; i<tmplist.size(); i++){
            (*frompy).push_back(tmplist[i].cast<int>());
          }
          (*(outputs[*outputsIDit].buffer))[count] = frompy;
        }
        else if(*outputsTypeit == "vector_float"){
          ////bp::list tmplist = bp::extract<bp::list>(d.get(*outputsNameit));
          ////RCPtr<Vector<float> > frompy(new Vector<float>());
          ////for(int i=0; i<len(tmplist); i++){
          ////  (*frompy).push_back(bp::extract<float>(tmplist[i]));
          ////}
          ////(*(outputs[*outputsIDit].buffer))[count] = frompy;
          py::list tmplist = d[(*outputsNameit).c_str()];
          RCPtr<Vector<float> > frompy(new Vector<float>());
          for(size_t i=0; i<tmplist.size(); i++){
            (*frompy).push_back(tmplist[i].cast<float>());
          }
          (*(outputs[*outputsIDit].buffer))[count] = frompy;
        }
        else if(*outputsTypeit == "vector_complex"){
          ////bp::list tmplist = bp::extract<bp::list>(d.get(*outputsNameit));
          ////RCPtr<Vector<complex<float> > > frompy(new Vector<complex<float> >());
          ////for(int i=0; i<len(tmplist); i++){
          ////  (*frompy).push_back(bp::extract<complex<float> >(tmplist[i]));
          ////}
          ////(*(outputs[*outputsIDit].buffer))[count] = frompy;
          py::list tmplist = d[(*outputsNameit).c_str()];
          RCPtr<Vector<std::complex<float> > > frompy(new Vector<std::complex<float> >());
          for(size_t i=0; i<tmplist.size(); i++){
            (*frompy).push_back(tmplist[i].cast<std::complex<float> >());
          }
          (*(outputs[*outputsIDit].buffer))[count] = frompy;
        }
        else if(*outputsTypeit == "matrix_int"){
          ////bp::list tmpmat = bp::extract<bp::list>(d.get(*outputsNameit));
          ////RCPtr<Matrix<int > > frompy 
          ////  (new Matrix<int >(bp::len(tmpmat), bp::len(tmpmat[0])));
          ////for(int r=0; r<bp::len(tmpmat); r++){
          ////  for(int c=0; c<bp::len(tmpmat[0]); c++){
          ////    (*frompy)(r, c) = bp::extract<int>(tmpmat[r][c]);
          ////  }
          ////}
          ////(*(outputs[*outputsIDit].buffer))[count] = frompy;
          py::list tmpmat = d[(*outputsNameit).c_str()];
          RCPtr<Matrix<int> > frompy(new Matrix<int>(tmpmat.size(), ((py::list)tmpmat[0]).size()));
          for(size_t r=0; r<tmpmat.size(); r++){
            for(size_t c=0; c<((py::list)tmpmat[0]).size(); c++){
              (*frompy)(r, c) = ((py::list)tmpmat[r])[c].cast<int>();
            }
          }
          (*(outputs[*outputsIDit].buffer))[count] = frompy;
        }  
        else if(*outputsTypeit == "matrix_float"){
          ////bp::list tmpmat = bp::extract<bp::list>(d.get(*outputsNameit));
          ////RCPtr<Matrix<float> > frompy (new Matrix<float>(len(tmpmat), len(tmpmat[0])));
          ////for(int r=0; r<bp::len(tmpmat); r++){
          ////  for(int c=0; c<bp::len(tmpmat[0]); c++){
          ////    (*frompy)(r, c) = bp::extract<float>(tmpmat[r][c]);
          ////  }
          ////}
          ////(*(outputs[*outputsIDit].buffer))[count] = frompy;
          py::list tmpmat = d[(*outputsNameit).c_str()];
          RCPtr<Matrix<float> > frompy(new Matrix<float>(tmpmat.size(), ((py::list)tmpmat[0]).size()));
          for(size_t r=0; r<tmpmat.size(); r++){
            for(size_t c=0; c<((py::list)tmpmat[0]).size(); c++){
              (*frompy)(r, c) = ((py::list)tmpmat[r])[c].cast<float>();
            }
          }
          (*(outputs[*outputsIDit].buffer))[count] = frompy;
        }
        else if(*outputsTypeit == "matrix_complex"){
          ////bp::list tmpmat = bp::extract<bp::list>(d.get(*outputsNameit));
          ////RCPtr<Matrix<complex<float> > > frompy (new Matrix<complex<float> >(len(tmpmat), len(tmpmat[0])));
          ////for(int r=0; r<len(tmpmat); r++){
          ////  for(int c=0; c<len(tmpmat[0]); c++){
          ////    (*frompy)(r, c) = bp::extract<complex<float> >(tmpmat[r][c]);
          ////  }
          ////}
          ////(*(outputs[*outputsIDit].buffer))[count] = frompy;
          py::list tmpmat = d[(*outputsNameit).c_str()];
          RCPtr<Matrix<std::complex<float> > > frompy(new Matrix<std::complex<float> >(tmpmat.size(), ((py::list)tmpmat[0]).size()));
          for(size_t r=0; r<tmpmat.size(); r++){
            for(size_t c=0; c<((py::list)tmpmat[0]).size(); c++){
              (*frompy)(r, c) = ((py::list)tmpmat[r])[c].cast<std::complex<float> >();
            }
          }
          (*(outputs[*outputsIDit].buffer))[count] = frompy;
        }
        else if(*outputsTypeit == "vector_source"){
          ////bp::list tmpvec = bp::extract<bp::list>(d.get(*outputsNameit));
          ////RCPtr<Vector<ObjectRef> > frompy (new Vector<ObjectRef>());
          ////for(int i=0; i<bp::len(tmpvec); i++){
          ////  RCPtr<Source> src = RCPtr<Source>(new Source);
          ////  bp::dict srcd = bp::extract<bp::dict>(tmpvec[i]);
          ////  src->id = bp::extract<int>(srcd.get("id"));
          ////  src->x[0] = bp::extract<float>(srcd.get("x")[0]);
          ////  src->x[1] = bp::extract<float>(srcd.get("x")[1]);
          ////  src->x[2] = bp::extract<float>(srcd.get("x")[2]);
          ////  src->power = bp::extract<float>(srcd.get("power"));
          ////  (*frompy).push_back(src);
          ////}
          ////(*(outputs[*outputsIDit].buffer))[count] = frompy;
          py::list tmpvec = d[(*outputsNameit).c_str()];
          RCPtr<Vector<ObjectRef> > frompy (new Vector<ObjectRef>());
          for(size_t i=0; i<tmpvec.size(); i++){
            RCPtr<Source> src = RCPtr<Source>(new Source);
            py::dict srcd = tmpvec[i];
            src->id = srcd["id"].cast<int>();
            src->x[0] = ((py::list)srcd["x"])[0].cast<float>();
            src->x[1] = ((py::list)srcd["x"])[1].cast<float>();
            src->x[2] = ((py::list)srcd["x"])[2].cast<float>();
            src->power = srcd["power"].cast<float>();
            (*frompy).push_back(src);
          }
          (*(outputs[*outputsIDit].buffer))[count] = frompy;
        }
        else if(*outputsTypeit == "map_vector_int"){
          ////bp::dict map = bp::extract<bp::dict>(d.get(*outputsNameit));
          ////bp::list keys = bp::extract<bp::list>(map.attr("keys")());
          ////RCPtr<Map<int, ObjectRef> > frompy (new Map<int, ObjectRef>);
          ////for(int i=0; i<bp::len(keys); i++){
          ////  RCPtr<Vector<int> > value = RCPtr<Vector<int> >(new Vector<int>);
          ////  bp::list tmplist = bp::extract<bp::list>(map.get(keys[i]));
          ////  for(int j=0; j<bp::len(tmplist); j++){
          ////    (*value).push_back(bp::extract<int>(tmplist[j]));
          ////  }
          ////  (*frompy)[bp::extract<int>(keys[i])] = value;
          ////}
          ////(*(outputs[*outputsIDit].buffer))[count] = frompy;
          py::dict map = d[(*outputsNameit).c_str()];
          py::list keys = map.attr("keys")();
          RCPtr<Map<int, ObjectRef> > frompy (new Map<int, ObjectRef>);
          for(size_t i=0; i<keys.size(); i++){
            RCPtr<Vector<int> > value = RCPtr<Vector<int> >(new Vector<int>);
            py::list tmplist = map[keys[i]];
            for(size_t j=0; j<tmplist.size(); j++){
              (*value).push_back(tmplist[j].cast<int>());
            }
            (*frompy)[keys[i].cast<int>()] = value;
          }
          (*(outputs[*outputsIDit].buffer))[count] = frompy;
        }
        else if(*outputsTypeit == "map_vector_float"){
          ////bp::dict map = bp::extract<bp::dict>(d.get(*outputsNameit));
          ////bp::list keys = bp::extract<bp::list>(map.attr("keys")());
          ////RCPtr<Map<int, ObjectRef> > frompy (new Map<int, ObjectRef>);
          ////for(int i=0; i<bp::len(keys); i++){
          ////  RCPtr<Vector<float> > value = RCPtr<Vector<float> >(new Vector<float>);
          ////  bp::list tmplist = bp::extract<bp::list>(map.get(keys[i]));
          ////  for(int j=0; j<bp::len(tmplist); j++){
          ////    (*value).push_back(bp::extract<float>(tmplist[j]));
          ////  }
          ////  (*frompy)[bp::extract<int>(keys[i])] = value;
          ////}
          ////(*(outputs[*outputsIDit].buffer))[count] = frompy;
          py::dict map = d[(*outputsNameit).c_str()];
          py::list keys = map.attr("keys")();
          RCPtr<Map<int, ObjectRef> > frompy (new Map<int, ObjectRef>);
          for(size_t i=0; i<keys.size(); i++){
            RCPtr<Vector<float> > value = RCPtr<Vector<float> >(new Vector<float>);
            py::list tmplist = map[keys[i]];
            for(size_t j=0; j<tmplist.size(); j++){
              (*value).push_back(tmplist[j].cast<float>());
            }
            (*frompy)[keys[i].cast<int>()] = value;
          }
          (*(outputs[*outputsIDit].buffer))[count] = frompy;
        }
        else if(*outputsTypeit == "map_vector_complex"){
          ////bp::dict map = bp::extract<bp::dict>(d.get(*outputsNameit));
          ////bp::list keys = bp::extract<bp::list>(map.attr("keys")());
          ////RCPtr<Map<int, ObjectRef> > frompy (new Map<int, ObjectRef>);
          ////for(int i=0; i<bp::len(keys); i++){
          ////  RCPtr<Vector<complex<float> > > value = RCPtr<Vector<complex<float> > >(new Vector<complex<float> >);
          ////  bp::list tmplist = bp::extract<bp::list>(map.get(keys[i]));
          ////  for(int j=0; j<bp::len(tmplist); j++){
          ////    (*value).push_back(bp::extract<complex<float> >(tmplist[j]));
          ////  }
          ////  (*frompy)[bp::extract<int>(keys[i])] = value;
          ////}
          ////(*(outputs[*outputsIDit].buffer))[count] = frompy;
          py::dict map = d[(*outputsNameit).c_str()];
          py::list keys = map.attr("keys")();
          RCPtr<Map<int, ObjectRef> > frompy(new Map<int, ObjectRef>);
          for(size_t i=0; i<keys.size(); i++){
            RCPtr<Vector<std::complex<float> > > value = RCPtr<Vector<std::complex<float> > >(new Vector<std::complex<float> >);
            py::list tmplist = map[keys[i]];
            for(size_t j=0; j<tmplist.size(); j++){
              (*value).push_back(tmplist[j].cast<std::complex<float> >());
            }
            (*frompy)[keys[i].cast<int>()] = value;
          }
          (*(outputs[*outputsIDit].buffer))[count] = frompy;
        }
        else if(*outputsTypeit == "map_prim_float"){
          ////bp::dict map = bp::extract<bp::dict>(d.get(*outputsNameit));
          ////bp::list keys = bp::extract<bp::list>(map.attr("keys")());
          ////RCPtr<Map<int, ObjectRef> > frompy (new Map<int, ObjectRef>);
          ////for(int i=0; i<bp::len(keys); i++){
          ////  RCPtr<Float> value = RCPtr<Float>(new Float);
          ////  (*frompy)[bp::extract<int>(keys[i])] = value;
          ////}
          ////(*(outputs[*outputsIDit].buffer))[count] = frompy;
          py::dict map = d[(*outputsNameit).c_str()];
          py::list keys = map.attr("keys")();
          RCPtr<Map<int, ObjectRef> > frompy (new Map<int, ObjectRef>);
          for(size_t i=0; i<keys.size(); i++){
            RCPtr<Float> value = RCPtr<Float>(new Float(map[keys[i]].cast<float>()));
            (*frompy)[keys[i].cast<int>()] = value;
          }
          (*(outputs[*outputsIDit].buffer))[count] = frompy;
        }
        else if(outputsTypeit->find("map_string_map_string:", 0) != std::string::npos){
          unsigned long long int index, offset;
          std::map<unsigned long long int, std::string> typesMap;
          index = outputsTypeit->find("map_string_map_string:", 0) + 22;
          offset = index;
          while(offset != std::string::npos){
            offset = outputsTypeit->find("prim_int", offset);
            if(offset != std::string::npos){
              typesMap[offset]="prim_int";
              offset+=8;
            }
            //std::cerr << "prim_int: count" << count << " offset" << offset << " npos" << std::string::npos << std::endl;
          }
          offset = index;
          while(offset != std::string::npos){
            offset = outputsTypeit->find("prim_float", offset);
            if(offset != std::string::npos){
              typesMap[offset]="prim_float";
              offset+=10;
            }
            //std::cerr << "prim_float: count" << count << " offset" << offset << " npos" << std::string::npos << std::endl;
          }
          std::map<unsigned long long int, std::string>::iterator it = typesMap.begin();
          //do{
          //  unsigned int mapend = outputsTypeit->find(";", index);
          //  index+=22;
          //  if(mapend == std::string::npos){}
          //  else{}
          //} while((index = outputsTypeit->find("map_string_map_string:", index)) != std::string::npos);
          ////bp::dict map = bp::extract<bp::dict>(d.get(*outputsNameit));
          ////bp::list keys = bp::extract<bp::list>(map.attr("keys")());
          ////keys.attr("sort")();
          ////RCPtr<Map<std::string, Map<std::string, ObjectRef> > > frompy = RCPtr<Map<std::string, Map<std::string, ObjectRef> > >(new Map<std::string, Map<std::string, ObjectRef> >);
          ////for(int i=0; i<bp::len(keys); i++){
          ////  RCPtr<Map<std::string, ObjectRef> > mapinmap = RCPtr<Map<std::string, ObjectRef> >(new Map<std::string, ObjectRef>);
          ////  bp::dict mapmap = bp::extract<bp::dict>(map.get(keys[i]));
          ////  bp::list mapkeys = bp::extract<bp::list>(mapmap.attr("keys")());
          ////  mapkeys.attr("sort")();
          ////  for(int j=0; j<bp::len(mapkeys); j++){
          ////    if(it->second == "prim_int"){
          ////      RCPtr<Int> value = RCPtr<Int>(new Int);
          ////      (*mapinmap)[bp::extract<std::string>(mapkeys[j])] = value;
          ////    }
          ////    else if(it->second == "prim_float"){
          ////      RCPtr<Float> value = RCPtr<Float>(new Float);
          ////      (*mapinmap)[bp::extract<std::string>(mapkeys[j])] = value;
          ////    }
          ////    it++;
          ////  }
          ////  (*frompy)[bp::extract<std::string>(keys[i])] = *mapinmap;
          ////}
          ////(*(outputs[*outputsIDit].buffer))[count] = frompy;
          py::dict map = d[(*outputsNameit).c_str()];
          py::list keys = map.attr("keys")();
          keys.attr("sort")();
          RCPtr<Map<std::string, Map<std::string, ObjectRef> > > frompy = RCPtr<Map<std::string, Map<std::string, ObjectRef> > >(new Map<std::string, Map<std::string, ObjectRef> >);
          for(size_t i=0; i<keys.size(); i++){
            RCPtr<Map<std::string, ObjectRef> > mapinmap = RCPtr<Map<std::string, ObjectRef> >(new Map<std::string, ObjectRef>);
            py::dict mapmap = map[keys[i]];
            py::list mapkeys = mapmap.attr("keys")();
            mapkeys.attr("sort")();
            for(size_t j=0; j<mapkeys.size(); j++){
              if(it->second == "prim_int"){
                RCPtr<Int> value = RCPtr<Int>(new Int);
                (*mapinmap)[mapkeys[j].cast<std::string>()] = value;
              }
              else if(it->second == "prim_float"){
                RCPtr<Float> value = RCPtr<Float>(new Float);
                (*mapinmap)[mapkeys[j].cast<std::string>()] = value;
              }
              it++;
            }
            (*frompy)[keys[i].cast<std::string>()] = *mapinmap;
          }
          (*(outputs[*outputsIDit].buffer))[count] = frompy;
        }
        else{
          std::stringstream ss;
          ss << "Unsupported type for output:" << *outputsTypeit;
          throw new NodeException(this, ss.str(), __FILE__, __LINE__);
        }
      }
    ////} catch(bp::error_already_set){
    } catch(const py::error_already_set& e){
      ////PyErr_Print();
      std::cerr << e.what() << std::endl; 
      std::stringstream ss;
      ss << "Error during setting outputs of type (" << *outputsTypeit << ") " << "and name (" << *outputsNameit << ")"
          << std::endl;
      throw (new NodeException(this, ss.str(), __FILE__, __LINE__));
    } catch(const std::exception& e) {
      std::cerr << e.what() << std::endl;
      std::stringstream ss;
      ss << "Error during setting outputs of type (" << *outputsTypeit << ") " << "and name (" << *outputsNameit << ")"
          << std::endl;
      throw (new NodeException(this, ss.str(), __FILE__, __LINE__));
    }

    isCalculated = true;
  }
};
